from . import arrdict
from torch import nn
from contextlib import contextmanager

class State:

    def __init__(self):
        super().__init__()

        self._value = None

    def get(self, factory=None):
        if self._value is None and factory is not None:
            self._value = factory()
        return self._value

    def set(self, value):
        self._value = value

    def clear(self):
        self._value = None

    def __repr__(self):
        return f'State({self._value})'
    
    def __str__(self):
        return repr(self)

def states(net):
    substates = {k: states(v) for k, v in net.named_children()}
    ownstates = {k: getattr(net, k) for k in dir(net) if isinstance(getattr(net, k), State)}
    return arrdict.arrdict({k: v for k, v in {**ownstates, **substates}.items() if v})

def _nonnull(x):
    y = type(x)()
    for k, v in x.items():
        if isinstance(v, dict):
            subtree = _nonnull(v)
            if subtree:
                y[k] = subtree
        elif v is not None:
            y[k] = v
    return y

def get(net):
    return _nonnull(states(net).map(lambda s: s.get()))

def set(net, state):
    state.starmap(lambda r, n: n.set(r), states(net))

def clear(net):
    states(net).map(lambda s: s.clear())

@contextmanager
def temp_clear(net):
    original = get(net)
    clear(net)
    try:
        yield
    finally:
        set(net, original)

@contextmanager
def temp_set(net, state):
    original = get(net)
    set(net, state)
    try:
        yield
    finally:
        set(net, original)

@contextmanager
def temp_clear_set(net, state):
    with temp_clear(net), temp_set(net, state):
        yield net

class Sequential(nn.Sequential):

    def forward(self, input, **kwargs):
        for module in self:
            input = module(input, **kwargs)
        return input


from contextlib import contextmanager
from functools import wraps

class MaybeAsyncGeneratorContextManager:

    def __init__(self, func, args, kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._sync = None
        self._async = None

    def __enter__(self):
        if self._sync is None:
            syncfunc = contextmanager(self._func)
            self._sync = syncfunc(*self._args, **self._kwargs)
        return self._sync.__enter__()

    def __exit__(self, t, v, tb):
        return self._sync.__exit__(t, v, tb)

    def __aenter__(self):
        if self._async is None:
            # Hide this 3.8 import; most users will never hit it
            from contextlib import asynccontextmanager

            @asynccontextmanager
            async def asyncfunc(*args, **kwargs):
                with contextmanager(self._func)(*args, **kwargs):
                    yield 
            self._async = asyncfunc(*self._args, **self._kwargs)
        return self._async.__aenter__()

    def __aexit__(self, t, v, tb):
        return self._async.__aexit__(t, v, tb)

def maybeasynccontextmanager(func):
    @wraps(func)
    def helper(*args, **kwds):
        return MaybeAsyncGeneratorContextManager(func, args, kwds)
    return helper


import numpy as np
import torch
import numbers
from rebar import arrdict
from . import cuda

class RaggedNumpy:

    def __init__(self, vals, widths):
        """A :ref:`Ragged <raggeds>` backed by numpy arrays.

        :param vals: a (V, ...)-array of backing values.
        :param widths: a (W,)-array of widths of each subarray in the ragged. The sum of the widths must equal ``V``.

        :var vals: a (V, ...)-array of backing values.
        :var widths: a (W,)-array of widths of each subarray in the ragged array.
        :var starts: a (W,)-array of indices giving where each subarray starts in ``vals``.
        :var ends: an (W,)-array of indices giving where each subarray ends in ``vals``.
        :var inverse: an (V,)-array of indices giving the index of the subarray the corresponding element of ``vals`` is a part of.
        """
        self.vals = vals
        self.widths = widths
        self.starts = widths.cumsum().astype(int) - widths
        self.ends = widths.cumsum().astype(int)

        assert widths.sum() == vals.shape[0]

        indices = np.zeros(self.widths.sum(), dtype=self.starts.dtype)
        indices[self.starts] = np.ones_like(self.starts)
        self.inverse = indices.cumsum().astype(int) - 1

    def __getitem__(self, x):
        """Indexes or slices the first dimension of the ragged array."""
        if isinstance(x, numbers.Integral):
            return self.vals[self.starts[x]:self.ends[x]]
        if isinstance(x, slice):
            assert x.step in (None, 1)
            start, end = x.start or 0, x.stop or len(self.ends) 
            return RaggedNumpy(
                self.vals[self.starts[start]:self.ends[end-1]],
                self.widths[start:end])
        raise ValueError(f'Can\'t handle index "{x}"')

    def torchify(self):
        """Applies :func:`arrdict.torchify` to the backing arrays and returns a new :class:`cuda.Ragged$ND` for them"""
        return Ragged(
            arrdict.torchify(self.vals),
            arrdict.torchify(self.widths))

    def __str__(self):
        return f'{type(self).__name__}({str(self.widths)})'

    def __repr__(self):
        return str(self)

def Ragged(vals, widths):
    """Returns a :ref:`Ragged <raggeds>` array or tensor. 
    
    If you pass numpy arrays as arguments, you'll get back a :class:`RaggedNumpy` object; if you pass Torch tensors,
    you'll get back a :class:`cuda.Ragged$ND` that's backed by a C++
    implementation and is OK to pass to the :class:`core.Core` machinery.

    :param vals: a (V, ...)-array/tensor of backing values.
    :param widths: a (W,)-array/tensor of widths of each subarray in the ragged. The sum of the widths must equal ``V``.

    :var vals: a (V, ...)-array/tensor of backing values.
    :var widths: a (W,)-array/tensor of widths of each subarray in the ragged.
    :var starts: a (W,)-array/tensor of indices giving where each subarray starts in ``vals``.
    :var ends: an (W,)-array/tensor of indices giving where each subarray ends in ``vals``.
    :var inverse: an (V,)-array/tensor of indices giving the index of the subarray the corresponding element of ``vals`` is a part of.
    """
    if isinstance(vals, np.ndarray):
        return RaggedNumpy(vals, widths)
    Ragged = getattr(cuda, f'Ragged{vals.ndim}D')
    return Ragged(vals, widths)

def test_ragged():
    vals = torch.as_tensor([0, 1, 2, 3, 4, 5]).float()
    widths = torch.as_tensor([3, 1, 2]).int()
    ragged = Ragged(vals, widths)

    torch.testing.assert_allclose(ragged[1], [3])
    torch.testing.assert_allclose(ragged[-1], [4, 5])

    torch.testing.assert_allclose(ragged[:2].vals, [0, 1, 2, 3])
    torch.testing.assert_allclose(ragged[:2].widths, [3, 1])

    torch.testing.assert_allclose(ragged[1:].vals, np.array([3, 4, 5]))
    torch.testing.assert_allclose(ragged[1:].widths, np.array([1, 2]))

def test_ragged_numpy():
    vals = np.array([0, 1, 2, 3, 4, 5])
    widths = np.array([3, 1, 2])
    ragged = RaggedNumpy(vals, widths)

    np.testing.assert_allclose(ragged[1], [3])
    np.testing.assert_allclose(ragged[-1], [4, 5])

    np.testing.assert_allclose(ragged[:2].vals, [0, 1, 2, 3])
    np.testing.assert_allclose(ragged[:2].widths, [3, 1])

    np.testing.assert_allclose(ragged[1:].vals, np.array([3, 4, 5]))
    np.testing.assert_allclose(ragged[1:].widths, np.array([1, 2]))



import pandas as pd
import pickle
from . import paths
import time

def store_latest(run_name, objs, throttle=0):
    path = paths.path(run_name, 'storing').with_suffix('.pkl')
    if path.exists():
        if (time.time() - path.lstat().st_mtime) < throttle:
            return False

    state_dicts = {k: v.state_dict() for k, v in objs.items()}
    bs = pickle.dumps(state_dicts)
    path.with_suffix('.tmp').write_bytes(bs)
    path.with_suffix('.tmp').rename(path)

    return True

def runs():
    return paths.runs()

def stored(run_name=-1):
    ps = paths.subdirectory(run_name, 'storing').glob('*.pkl')
    infos = []
    for p in ps:
        infos.append({
            **paths.parse(p),
            'path': p})

    return pd.DataFrame(infos)

def load(run_name=-1, procname='MainProcess'):
    path = stored(run_name).loc[lambda df: df.procname == procname].iloc[-1].path
    return pickle.loads(path.read_bytes())


"""TODO-DOCS Module docs""" 

import numpy as np
import torch
from rebar import arrdict
from . import spaces, geometry, cuda, plotting
import matplotlib.pyplot as plt

def to_local_frame(angles, p):
    a = np.pi/180*angles
    c, s = torch.cos(a), torch.sin(a)
    x, y = p[..., 0], p[..., 1]
    return torch.stack([c*x + s*y, -s*x + c*y], -1)

def to_global_frame(angles, p):
    a = np.pi/180*angles
    c, s = torch.cos(a), torch.sin(a)
    x, y = p[..., 0], p[..., 1]
    return torch.stack([c*x - s*y, s*x + c*y], -1)

class SimpleMovement:

    def __init__(self, core, *args, accel=10, ang_accel=180, n_agents=None, **kwargs):
        # noop, forward/backward, strafe left/right, turn left/right
        self.core = core
        velocity = torch.tensor([[0., 0.], [0., 1.], [0.,-1.], [1., 0.], [-1.,0.], [0., 0.], [0., 0.]])
        angvelocity = torch.tensor([0., 0., 0., 0., 0., +1., -1.])
        self._actionset = arrdict.arrdict(
            velocity=accel/core.fps*velocity,
            angvelocity=ang_accel/core.fps*angvelocity
        ).to(core.device)

        self.space = spaces.MultiDiscrete(n_agents or core.n_agents, 7)

    def __call__(self, decision):
        core = self.core
        delta = self._actionset[decision.actions.long()]
        core.agents.angvelocity[:] = delta.angvelocity
        core.agents.velocity[:] = to_global_frame(core.agents.angles, delta.velocity)
        return cuda.physics(core.scenery, core.agents)

class MomentumMovement:

    def __init__(self, core, *args, accel=5, ang_accel=180, decay=.125, n_agents=None, **kwargs):
        # noop, forward/backward, strafe left/right, turn left/right
        self.core = core
        velocity = torch.tensor([[0., 0.], [0., 1.], [0.,-1.], [1., 0.], [-1.,0.], [0., 0.], [0., 0.]])
        angvelocity = torch.tensor([0., 0., 0., 0., 0., +1., -1.])
        self._actionset = arrdict.arrdict(
            velocity=accel/core.fps*velocity,
            angvelocity=ang_accel/core.fps*angvelocity
        ).to(core.device)

        self._decay = decay

        self.space = spaces.MultiDiscrete(n_agents or core.n_agents, 7)

    def __call__(self, decision):
        core = self.core
        delta = self._actionset[decision.actions.long()]
        core.agents.angvelocity[:] = (1 - self._decay)*core.agents.angvelocity + delta.angvelocity
        core.agents.velocity[:] = (1 - self._decay)*core.agents.velocity + to_global_frame(core.agents.angles, delta.velocity)
        return cuda.physics(core.scenery, core.agents)

def unpack(d):
    if isinstance(d, torch.Tensor):
        return d
    return arrdict.arrdict({k: unpack(getattr(d, k)) for k in dir(d) if not k.startswith('_')})

def render(core):
    render = unpack(cuda.render(core.scenery, core.agents))
    render = arrdict.arrdict({k: v.unsqueeze(2) for k, v in render.items()})
    render['screen'] = render.screen.permute(0, 1, 4, 2, 3)
    return render

def downsample(screen, subsample):
    return screen.view(*screen.shape[:-1], screen.shape[-1]//subsample, subsample)

class Depth:

    def __init__(self, core, *args, n_agents=None, subsample=1, max_depth=10, **kwargs):
        n_agents = n_agents or core.n_agents
        self.core = core
        self.space = spaces.MultiImage(n_agents, 1, 1, core.res//subsample)
        self.max_depth = max_depth
        self.subsample = subsample

    def __call__(self, r=None):
        r = render(self.core) if r is None else r
        depth = ((r.distances - self.core.agent_radius)/self.max_depth).clamp(0, 1)
        self._last_obs = downsample(depth, self.subsample).mean(-1).unsqueeze(3)
        return self._last_obs
    
    def state(self, e):
        return self._last_obs[e].clone()

class RGB:

    def __init__(self, core, *args, n_agents=None, subsample=1, **kwargs):
        n_agents = n_agents or core.n_agents
        self.core = core
        self.space = spaces.MultiImage(n_agents, 3, 1, core.res//subsample)
        self.subsample = subsample

    def __call__(self, r=None):
        r = render(self.core) if r is None else r
        self._last_obs = downsample(r.screen, self.subsample).mean(-1)
        return self._last_obs
    
    def state(self, e):
        return self._last_obs[e].clone()

    @classmethod
    def plot_state(cls, state, axes=None):
        n_agents = state.shape[0]
        axes = plt.subplots(n_agents, 1, squeeze=False) if axes is None else axes
        plotting.plot_images({'rgb': state}, axes)
        return axes

class IMU:

    def __init__(self, core, n_agents=None):
        self.core = core
        self.space = spaces.MultiVector(n_agents or core.n_agents, 3)

    def __call__(self):
        return torch.cat([
            self.core.agents.angvelocity[..., None]/360.,
            to_local_frame(self.core.agents.angles, self.core.agents.velocity)/10.], -1)

def random_empty_positions(geometries, n_agents, n_points):
    points = []
    for g in geometries:
        sample = np.stack((g.masks > 0).nonzero(), -1)

        # There might be fewer open points than we're asking for
        n_possible = min(len(sample)//n_agents, n_points)
        sample = sample[np.random.choice(np.arange(len(sample)), (n_possible, n_agents), replace=True)]

        # So repeat the sample until we've got enough
        sample = np.concatenate([sample]*int(n_points/len(sample)+1))[-n_points:]
        sample = np.random.permutation(sample)
        points.append(geometry.centers(sample, g.masks.shape, g.res).transpose(1, 0, 2))
    return arrdict.stack(points)
        
class RandomSpawns:

    def __init__(self, geometries, core, *args, n_spawns=100, **kwargs):
        self.core = core

        positions = random_empty_positions(geometries, core.n_agents, n_spawns)
        angles = core.random.uniform(-180, +180, (len(geometries), core.n_agents, n_spawns))
        self._spawns = arrdict.torchify(arrdict.arrdict(positions=positions, angles=angles)).to(core.device)

    def __call__(self, reset):
        core = self.core
        required = reset.nonzero(as_tuple=True)
        choices = torch.randint_like(required[0], 0, self._spawns.angles.shape[1])
        core.agents.angles[required] = self._spawns.angles[(*required, choices)] 
        core.agents.positions[required] = self._spawns.positions[(*required, choices)] 
        core.agents.velocity[required] = 0.
        core.agents.angvelocity[required] = 0.

class RandomLengths:

    def __init__(self, core, max_length=512, min_length=None):
        min_length = max_length//2 if min_length is None else min_length
        self._max_lengths = torch.randint(min_length, max_length, (core.n_envs,), dtype=torch.int, device=core.device)
        self._lengths = torch.zeros_like(self._max_lengths)
    
    def __call__(self, reset=None):
        self._lengths += 1
        reset = torch.zeros_like(self._lengths, dtype=torch.bool) if reset is None else reset
        reset = (self._lengths >= self._max_lengths) | reset
        self._lengths[reset] = 0
        return reset

    def state(self, d):
        return arrdict.arrdict(length=self._lengths[d], max_length=self._max_lengths[d]).clone()



import os
from pathlib import Path
import shutil
import multiprocessing as mp
import pandas as pd
import re
from . import dotdict

ROOT = 'output/traces'

def resolve(run_name):
    if isinstance(run_name, str):
        return run_name
    if isinstance(run_name, int):
        times = {p: p.stat().st_ctime for p in Path(ROOT).iterdir()}
        paths = sorted(times, key=times.__getitem__)
        return paths[run_name].parts[-1]
    raise ValueError(f'Can\'t find a run corresponding to {run_name}')

def run_dir(run_name):
    run_name = resolve(run_name)
    return Path(ROOT) / run_name

def subdirectory(run_name, group, channel=''):
    if channel:
        return run_dir(run_name) / group / channel
    else:
        return run_dir(run_name) / group 

def clear(run_name, group=None):
    if group is None:
        shutil.rmtree(run_dir(run_name), ignore_errors=True)
    else:
        shutil.rmtree(subdirectory(run_name, group), ignore_errors=True)

def path(run_name, group, channel=''):
    # Python's idea of a process name is different from the system's idea. Dunno where
    # the difference comes from.
    run_name = resolve(run_name)

    proc = mp.current_process()

    for x in [run_name, group]:
        for c in ['_', os.sep]:
            assert c not in x, f'Can\'t have "{c}" in the file path'

    path = subdirectory(run_name, group, channel) / f'{proc.name}-{proc.pid}'

    path.parent.mkdir(exist_ok=True, parents=True)
    return path

def glob(run_name, group, channel='', pattern='*'):
    paths = subdirectory(run_name, group, channel).glob(pattern)
    return sorted(paths, key=lambda p: p.stat().st_mtime)

def parse(path):
    parts = path.relative_to(ROOT).with_suffix('').parts
    procname, pid = re.match(r'^(.*)-(.*)$', parts[-1]).groups()
    return dotdict.dotdict(
        run_name=parts[0], 
        group=parts[1], 
        channel='/'.join(parts[2:-1]),
        filename=parts[-1],
        procname=procname, 
        pid=pid)

def runs():
    paths = []
    for p in Path(ROOT).iterdir():
        paths.append({
            'path': p, 
            'created': pd.Timestamp(p.stat().st_ctime, unit='s'),
            'run_name': p.parts[-1]})
    return pd.DataFrame(paths).sort_values('created').reset_index(drop=True)

def size(run_name, group):
    run_name = resolve(run_name)
    b = sum(item.stat().st_size for item in subdirectory(run_name, group).glob('**/*.*'))
    return b/1e6



import numpy as np
from . import geometry
from rebar import arrdict

def box(width=5):
    """A :ref:`geometry <geometry>` which is just a simple box, with one room and one light inside it."""
    corners = [(np.cos(t), np.sin(t)) for t in np.arange(np.pi/4, 2*np.pi, np.pi/2)]
    corners = width/2**.5 *np.array(corners) + width/2 + geometry.MARGIN
    walls = np.stack(geometry.cyclic_pairs(corners))
    spaces = [corners]

    return arrdict.arrdict(
        walls=walls,
        lights=np.full((1, 2), width/2 + geometry.MARGIN),
        masks=geometry.masks(walls, spaces),
        res=geometry.RES)

def column(width=5, column_width=.1):
    """A :ref:`geometry <geometry>` which is just a simple 'column' (aka small box), with one room around it"""
    corners = [(np.cos(t), np.sin(t)) for t in np.arange(np.pi/4, 2*np.pi, np.pi/2)]
    column_corners = column_width/2**.5 *np.array(corners) + width/2 + geometry.MARGIN
    walls = np.stack(geometry.cyclic_pairs(column_corners))
    spaces = [width/2**.5 * np.array(corners) + width/2 + geometry.MARGIN]

    return arrdict.arrdict(
        walls=walls,
        lights=2**.5 * np.array(corners) + width/2 + geometry.MARGIN,
        masks=geometry.masks(walls, spaces),
        res=geometry.RES)



"""TODO-DOCS Geometry docs"""
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import rasterio.features
from shapely.ops import cascaded_union
from shapely.geometry import Polygon, LineString
from bs4 import BeautifulSoup

MARGIN = 1.
RES = .2
SCALE = 100

from itertools import islice, cycle

def cyclic_pairs(xs):
    """Returns pairs ``(xs[i], xs[i+1])``, wrapping the last pair round to the start."""
    ys = islice(cycle(xs), 1, None)
    return list(zip(xs, ys))

def signed_area(points):
    area = 0
    for x, y in cyclic_pairs(points):
        area += x[0]*y[1] - x[1]*y[0]
    return area

def polypoints(svgpoly):
    return np.array([list(map(float, p.split(','))) for p in svgpoly.attrs['points'].split()])

def orient(points):
    if signed_area(points) > 0:
        return points
    else:
        return points[::-1]

def unique(walls):
    """Eliminate walls that are copies of other walls"""
    forward  = ((walls[:, None, :, :] - walls[None, :, ::+1, :])**2).sum(-1).sum(-1)**.5
    backward = ((walls[:, None, :, :] - walls[None, :, ::-1, :])**2).sum(-1).sum(-1)**.5
    mask = (forward < 1e-3) | (backward < 1e-3)
    mask[np.triu_indices_from(mask)] = False
    return walls[~mask.any(1)]

def svg_walls(soup):
    walls = cascaded_union([Polygon(polypoints(wall)) for wall in soup.select('.Wall>polygon')])
    doors = cascaded_union([Polygon(polypoints(door)) for door in soup.select('.Door>polygon')])

    # Sometimes the doors are slightly misaligned with the walls; dilating the doors handles this
    skeleton = walls - doors.buffer(5)
    if skeleton.geom_type == 'MultiPolygon':
        tops = [orient(np.array(g.exterior.coords)) for g in skeleton.geoms]
    elif skeleton.geom_type == 'Polygon':
        tops = [orient(np.array(skeleton.exterior.coords))]
    walls = np.concatenate([cyclic_pairs(t) for t in tops])

    # Zero-length walls cause various warnings upstream. May as well get rid of them now. 
    lengths = ((walls[:, 0] - walls[:, 1])**2).sum(1)**.5
    return unique(walls[lengths > 0])

def svg_spaces(soup):
    return [polypoints(poly) for poly in soup.select('.Space>polygon')]

def transform(walls, spaces):
    """svg coords are in centimeters from the (left, top) corner, 
    while we want metres from the (left, bottom) corner"""
    joint = np.concatenate([np.concatenate(walls), np.concatenate(spaces)])
    (left, _), (_, bot) = joint.min(0), joint.max(0)

    def tr(ps):
        x, y = ps[..., 0], ps[..., 1]
        return np.stack([x - left, bot - y], -1)/SCALE + MARGIN
    
    return tr(walls), [tr(s) for s in spaces]

def mask_transform(*args):
    points = np.concatenate([np.concatenate(a) for a in args])
    assert np.concatenate(points).min() > 0, 'Masker currently requires the points to be in the top-right quadrant'
    r, t = points.max(0) + MARGIN
    h, w = int(t/RES)+1, int(r/RES)+1
    return rasterio.transform.Affine(RES, 0, 0, 0, -RES, h*RES), (h, w)

def masks(walls, spaces, res=RES):
    """Generates a masking array from an array of walls and a list of spaces.
    
    :param walls: A (n_walls, 2, 2)-array giving the coordinates of the walls' endpoints.
    :param spaces: A list of spaces, each given as a coordinate array of the space's vertices.
    :param res: The resolution of the the masking array. 
    :return: A masking array, with indices 1, 2, ... for the spaces, 0 for free space, and -1 for walls. 
    """
    transform, shape = mask_transform(walls, spaces)
    wall_shapes = [(cascaded_union([LineString(p).buffer(.01) for p in walls]), -1)]
    space_shapes = [(Polygon(p).buffer(0), i+1) for i, p in enumerate(spaces)]
    shapes = [(s, v) for (s, v) in space_shapes + wall_shapes if not s.is_empty]
    return rasterio.features.rasterize(shapes, shape, transform=transform, all_touched=True, dtype=np.int16)

def centroids(spaces):
    # Reshape needed for the case there are no lights
    return np.array([Polygon(ps).centroid.coords[0] for ps in spaces]).reshape(-1, 2)

def geometry(svg):
    soup = BeautifulSoup(svg, features='lxml')
    walls = unique(svg_walls(soup))
    spaces = svg_spaces(soup)
    walls, spaces = transform(walls, spaces)
    return dict(
        walls=walls,
        lights=centroids(spaces),
        masks=masks(walls, spaces),
        res=RES)

def centers(indices, shape, res):
    """Converts mask (i, j) indices to the (x, y) coordinates of the ``ij`` th cell's center.

    Usually the ``shape`` and ``res`` arguments for this come directly from a geoemtry dotdict.
    
    :param indices: A (..., 2) array of indices into a masking array.
    :param shape: A tuple-like giving the height and width of the masking array.
    :param res: The resolution of the masking array.
    :return: A (..., 2) array of (x, y) coordinates
    """
    i, j = indices[..., 0] + .5, indices[..., 1] + .5
    xy = res*np.stack([j, shape[0] - i], -1)
    return xy

def indices(coords, shape, res):
    """Converts (x, y) coordinates to the (i, j) indices of the containing cell.
    
    Usually the ``shape`` and ``res`` arguments for this come directly from a geoemtry dotdict.
    
    :param indices: A (..., 2) array of (x, y) coordinates.
    :param shape: A tuple-like giving the height and width of the masking array.
    :param res: The resolution of the masking array.
    :return: A (..., 2) array of integer (i, j) indices.
    """
    x, y = coords[..., 0], coords[..., 1]
    i = (shape[0] - y/res).clip(0, shape[0]-1)
    j = (x/res).clip(0, shape[1]-1)
    return np.stack([i, j], -1).astype(int)

def display(g):
    """Visualize a geometry using matplotlib. 
    
    Supports visualizing partial geometries, that only have a subset of id/masks/walls/lights"""
    fig, ax = plt.subplots()
    ax.set_aspect(1)

    if 'id' in g:
        ax.set_title(g.id)

    if 'walls' in g:
        color = 'k' if 'masks' in g else 'k'
        lines = mpl.collections.LineCollection(g.walls, color=color, linewidth=2)
        ax.add_collection(lines)
        ax.autoscale()

    if 'lights' in g:
        for light in g.lights:
            ax.add_patch(mpl.patches.Circle(light[:2], radius=.05, color='yellow'))
        ax.autoscale()

    if 'masks' in g:
        height, width = g.res*np.array(g.masks.shape)
        extent = (0, width, 0, height)
        cm = ax.imshow(g.masks, extent=extent, cmap='tab20')
        ticks = np.arange(g.masks.min(), g.masks.max()+1)
        plt.colorbar(cm, values=ticks, ticks=ticks)
    
    return fig


from multiprocessing import Value
import numpy as np
from functools import partialmethod
from . import dotdict
try:
    import torch
    TORCH = True
except ModuleNotFoundError:
    TORCH = False

def _arrdict_factory():
    # This is done with a factory because I am a lazy man and I didn't fancy defining all the binary ops on 
    # the arrdict manually.

    class _arrdict_base(dotdict.dotdict):
        """An arrdict is an :class:`~rebar.dotdict.dotdict` with extra support for array and tensor values.

        arrdicts have a lot of unusual but extremely useful behaviours, which are documented in :ref:`the dotdicts
        and arrdicts concept section <dotdicts>` .
        """

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __getitem__(self, x):
            if isinstance(x, str):
                return super().__getitem__(x)
            return type(self)({k: v[x] for k, v in self.items()})

        def __binary_op__(self, name, rhs):
            if isinstance(rhs, dict):
                return self.starmap(name, rhs)
            else:
                return super().__getattr__(name)(rhs)

    # Add binary methods
    # https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types
    binaries = [
        'lt', 'le', 'eq', 'ne', 'ge', 'gt', 
        'add', 'sub', 'mul', 'matmul', 'truediv', 'floordiv', 'mod', 'divmod', 'pow', 'lshift', 'rshift', 'and', 'or', 'xor',
        'radd', 'rsub', 'rmul', 'rmatmul', 'rtruediv', 'rfloordiv', 'rmod', 'rdivmod', 'rpow', 'rand', 'lshift', 'rshift', 'ror', 'rxor']
    methods = {f'__{name}__': partialmethod(_arrdict_base.__binary_op__, f'__{name}__') for name in binaries}

    methods['__doc__'] = _arrdict_base.__doc__

    return type('arrdict', (_arrdict_base,), methods)

arrdict = _arrdict_factory()

@dotdict.mapping
def torchify(a):
    """Converts an array or a dict of numpy arrays to CPU tensors.

    If you'd like CUDA tensors, follow the tensor-ification ``.cuda()`` ; the attribute delegation
    built into :class:`~rebar.dotdict.dotdict` s will do the rest.
    
    Floats get mapped to 32-bit PyTorch floats; ints get mapped to 32-bit PyTorch ints. This is usually what you want in 
    machine learning work.
    """
    if hasattr(a, 'torchify'):
        return a.torchify()

    a = np.asarray(a)
    if np.issubdtype(a.dtype, np.floating):
        dtype = torch.float
    elif np.issubdtype(a.dtype, np.integer):
        dtype = torch.int
    elif np.issubdtype(a.dtype, np.bool_):
        dtype = torch.bool
    else:
        raise ValueError(f'Can\'t handle {type(a)}')
    return torch.as_tensor(np.array(a), dtype=dtype)

@dotdict.mapping
def numpyify(tensors):
    """Converts an array or a dict of tensors to numpy arrays.
    """
    if isinstance(tensors, tuple):
        return tuple(numpyify(t) for t in tensors)
    if isinstance(tensors, torch.Tensor):
        return tensors.clone().detach().cpu().numpy()
    if hasattr(tensors, 'numpyify'):
        return tensors.numpyify()
    return tensors

def stack(x, *args, **kwargs):
    """Stacks a sequence of arrays, tensors or dicts thereof.  

    For example, 

    >>> d = arrdict(a=1, b=np.array([1, 2]))
    >>> stack([d, d, d])
    arrdict:
    a    ndarray((3,), int64)
    b    ndarray((3, 2), int64)

    Any ``*args`` or ``**kwargs`` will be forwarded to the ``np.stack`` or ``torch.stack`` call. 

    Python scalars are converted to numpy scalars, so - as in the example above - stacking floats will
    get you a 1D array.
    """
    if isinstance(x[0], dict):
        ks = x[0].keys()
        return x[0].__class__({k: stack([y[k] for y in x], *args, **kwargs) for k in ks})
    if TORCH and isinstance(x[0], torch.Tensor):
        return torch.stack(x, *args, **kwargs)
    if isinstance(x[0], np.ndarray):
        return np.stack(x, *args, **kwargs) 
    if np.isscalar(x[0]):
        return np.array(x, *args, **kwargs)
    raise ValueError(f'Can\'t stack {type(x[0])}')

def cat(x, *args, **kwargs):
    """Concatenates a sequence of arrays, tensors or dicts thereof.  

    For example, 

    >>> d = arrdict(a=1, b=np.array([1, 2]))
    >>> cat([d, d, d])
    arrdict:
    a    ndarray((3,), int64)
    b    ndarray((6,), int64)

    Any ``*args`` or ``**kwargs`` will be forwarded to the ``np.concatenate`` or ``torch.cat`` call. 

    Python scalars are converted to numpy scalars, so - as in the example above - concatenating floats will
    get you a 1D array. 
    """
    if isinstance(x[0], dict):
        ks = x[0].keys()
        return x[0].__class__({k: cat([y[k] for y in x], *args, **kwargs) for k in ks})
    if TORCH and isinstance(x[0], torch.Tensor):
        return torch.cat(x, *args, **kwargs)
    if isinstance(x[0], np.ndarray):
        return np.concatenate(x, *args, **kwargs) 
    if np.isscalar(x[0]):
        return np.array(x)
    raise ValueError(f'Can\'t cat {type(x[0])}')

@dotdict.mapping
def clone(t):
    if hasattr(t, 'clone'):
        return t.clone()
    if hasattr(t, 'copy'):
        return t.copy()
    return t


import torch
import numpy as np
import inspect
from ..contextlib import maybeasynccontextmanager
from .. import numpy
from . import categories
from functools import partial

__all__ = ['to_dir', 'defer', 'record']

WRITER = None

@maybeasynccontextmanager
def to_dir(run_name):
    try:
        global WRITER
        old = WRITER
        WRITER = numpy.Writer(run_name, 'stats')
        yield
    finally:
        WRITER = old
def clean(x):
    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    if isinstance(x, np.ndarray) and x.ndim == 0:
        x = x.item()
    if isinstance(x, dict):
        return {k: clean(v) for k, v in x.items()}
    return x

def eager_record(category, field, *args, **kwargs):
    if WRITER is None:
        return 
    if not isinstance(field, str):
        raise ValueError(f'Field should be a string, is actually {field}')

    args = tuple(clean(a) for a in args)
    kwargs = {k: clean(v) for k, v in kwargs.items()}

    func = categories.CATEGORIES[category]
    call = inspect.getcallargs(func, *args, **kwargs)
    call = {'_time': np.datetime64('now'), **call}

    WRITER.write(f'{category}/{field}', call)

_record = eager_record
QUEUE = None

def record(*args, **kwargs):
    return _record(*args, **kwargs)
def deferred_record(category, field, *args, **kwargs):
    if not isinstance(field, str):
        raise ValueError(f'Field should be a string, is actually {field}')
    QUEUE.append((category, field, args, kwargs))

def _mono_getter(collection, x):
    dtype = x.dtype
    if dtype not in collection:
        collection[dtype] = []
    start = sum(c.nelement() for c in collection[dtype])
    end = start + x.nelement()
    collection[dtype].append(x.flatten())

    def f(collection):
        return collection[dtype][start:end].reshape(x.shape)
    return f

def _dummy_getter(x):
    def f(collection):
        return x
    return f

def _multi_getter(collection, *args, **kwargs):
    arggetters = []
    for a in args:
        if isinstance(a, torch.Tensor) and a.device.type != 'cpu':
            arggetters.append(_mono_getter(collection, a))
        else:
            arggetters.append(_dummy_getter(a))

    kwarggetters = {}
    for k, v in kwargs.items():
        if isinstance(v, torch.Tensor) and v.device.type != 'cpu':
            kwarggetters[k] = _mono_getter(collection, v)
        else:
            kwarggetters[k] = _dummy_getter(v)

    def f(collection):
        args = tuple(g(collection) for g in arggetters)
        kwargs = {k: g(collection) for k, g in kwarggetters.items()}
        return args, kwargs
    return f

def _gather(queue):
    collection = {}
    getters = []
    for category, field, args, kwargs in queue:
        getters.append((category, field, _multi_getter(collection, *args, **kwargs)))
    collection = {k: torch.cat(v).detach().cpu() for k, v in collection.items()}
    return collection, getters

@maybeasynccontextmanager
def defer():
    global _record
    global QUEUE
    _record = deferred_record
    QUEUE = []
    try:
        yield
    finally:
        collection, getters = _gather(QUEUE)

        for (category, field, getter) in getters:
            args, kwargs = getter(collection)
            args = tuple(clean(a) for a in args)
            kwargs = {k: clean(v) for k, v in kwargs.items()}
            func = categories.CATEGORIES[category]
            call = inspect.getcallargs(func, *args, **kwargs)
            call = {'_time': np.datetime64('now'), **call}

            if WRITER is not None:
                WRITER.write(f'{category}/{field}', call)
    
        QUEUE = None
        _record = eager_record

for c in categories.CATEGORIES:
    locals()[c] = partial(record, c)
    __all__.append(c)


import numpy as np
from . import scene, cuda, ragged
import torch
import logging
from rebar import arrdict, dotdict
import matplotlib.pyplot as plt

log = logging.getLogger(__name__)

AGENT_WIDTH = .15
TEXTURE_RES = .05

# Used for collision radius and near camera plane
AGENT_RADIUS = 1/2**.5*AGENT_WIDTH

def gamma_encode(x): 
    """Converts to viewable values"""
    return x**(1/2.2)

def gamma_decode(x):
    """Converts to interpolatable values"""
    return x**2.2

def _init_agents(n_envs, n_agents, device='cuda'):
    """Creates and returns an Agents datastructure"""
    data = arrdict.arrdict(
            angles=torch.zeros((n_envs, n_agents)),
            positions=torch.zeros((n_envs, n_agents, 2)),
            angvelocity=torch.zeros((n_envs, n_agents)),
            velocity=torch.zeros((n_envs, n_agents, 2)))
    return cuda.Agents(**data.to(device))

class Core: 

    def __init__(self, scenery, res=64, fov=130, fps=10):
        """The core rendering and physics interface. 

        To create the Core, you pass a :class:`~megastep.cuda.Scenery` that describes the environment. Once created, 
        the tensors hanging off of the Core give the state of the world, and that state can be advanced with the 
        functions in :mod:`~megastep.cuda`.

        :var agents: A :class:`~megastep.cuda.Agents` object describing the agents.
        :var scenery: A :class:`~megastep.cuda.Scenery` object describing the scene.
        :var progress: A (n_env, n_agent)-tensor giving how far the agent was able to move in the previous timestep as a 
            fraction of its intended movement, before running into an obstable. A value less than 1 means the agent collided
            with something. Useful for detecting collisions.
        :var n_envs: Number of environments. Same as the number of geometries passed in.
        :var n_agents: Number of agents.
        :var res: The horizontal resolution of observations. 
        :var fov: The field of view in degrees.
        :var agent_radius: The radius of a disc containing the agent, in meters.
        :var fps: The framerate.
        :var random: The seeded :class:`numpy.random.RandomState` used to initialize the environment. By reusing this
            in any extra random decisions made when generating the environments, it can be guaranteed you'll get the same
            environments every time.

        :param scenery: Describes the static parts of the environment.
        :type scenery: :class:`~megastep.cuda.Scenery`.
        :param n_agents: the number of agents to put in each environment. Defaults to 1.
        :type n_agents: int
        :param res: The horizontal resolution of the observations. The resolution must be less than 1024, as
            that's the `maximum number of CUDA threads in a block
            <https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#compute-capabilities>`_. Defaults to 64
            pixels.
        :type res: int
        :param fov: The field of view in degrees. Must be less than 180° due to how frames are rendered. Defaults to 130°.
        :type fov float:
        :param fps: The simulation frame rate/step rate. Defaults to 10. 
        :type fps: int
        """

        self.n_envs = len(scenery.lines.widths)
        self.n_agents = scenery.n_agents
        self.res = res
        self.fov = fov
        # TODO: Make this adjustable
        self.agent_radius = AGENT_RADIUS
        self.fps = fps
        self.random = np.random.RandomState(1)

        # TODO: This needs to be propagated to the C++ side
        self.device = scenery.model.device

        assert fov < 180, 'FOV should be less than 180°'

        cuda.initialize(self.agent_radius, self.res, self.fov, self.fps)
        self.scenery = scenery 
        self.agents = _init_agents(self.n_envs, self.n_agents, self.device)
        self.progress = torch.ones((self.n_envs, self.n_agents), device=self.device)

    def state(self, e):
        """Returns a :class:`~rebar.dotdict.dotdict` tree representing the state of environment ``e``.

        A typical state looks like this::

            arrdict:
            n_envs          1
            n_agents        4
            res             512
            fov             60
            agent_radius    0.10606601717798211
            fps             10
            scenery           arrdict:
                            model       Tensor((8, 2, 2), torch.float32)
                            lines       Tensor((307, 2, 2), torch.float32)
                            lights      Tensor((21, 3), torch.float32)
                            textures    <megastepcuda.Ragged2D object at 0x7fba34112eb0>
                            baked       <megastepcuda.Ragged1D object at 0x7fba34112670>
            agents          arrdict:
                            angles       Tensor((4,), torch.float32)
                            positions    Tensor((4, 2), torch.float32)
            progress        Tensor((4,), torch.float32)

        This state tree is usually passed onto a :ref:`plotting` function.""" 
        options = ('n_envs', 'n_agents', 'res', 'fov', 'agent_radius', 'fps')
        options = {k: getattr(self, k) for k in options}

        return arrdict.clone(dotdict.dotdict(
                **options,
                scenery=self.scenery.state(e),
                agents=self.agents.state(e),
                progress=self.progress[e]))
    
    @classmethod
    def plot_state(cls, state, ax=None, zoom=False):
        from . import plotting
        ax = ax or plt.axes()
        plotting.plot_lines(ax, state, zoom=zoom)
        plotting.plot_lights(ax,state)
        plotting.adjust_view(ax, state, zoom=zoom)
        plotting.plot_fov(ax, state)
        ax.set_xticks([])
        ax.set_yticks([])
        return ax

    def env_full(self, x):
        """Returns a (n_envs,)-tensor on the environment's device full of `x`.

        This isn't strictly required by the Core, but you find yourself making these vectors so often it's useful sugar.
        """
        dtypes = {bool: torch.bool, int: torch.int32, float: torch.float32}
        return torch.full((self.n_envs,), x, device=self.device, dtype=dtypes[type(x)])

    def agent_full(self, x):
        """Returns a (n_envs, n_agents)-tensor on the environment's device full of `x`.
        
        This isn't strictly required by the Core, but you find yourself making these vectors so often it's useful sugar.
        """
        dtypes = {bool: torch.bool, int: torch.int32, float: torch.float32}
        return torch.full((self.n_envs, self.n_agents), x, device=self.device, dtype=dtypes[type(x)])


import time
import torch
from .. import numpy, paths, widgets, logging
import re
import numpy as np
from .. import arrdict
from . import categories
import pandas as pd
import threading
from contextlib import contextmanager
import _thread

log = logging.getLogger(__name__)

def format(v):
    if isinstance(v, int):
        return f'{v}'
    if isinstance(v, float):
        return f'{v:.6g}'
    if isinstance(v, list):
        return ', '.join(format(vv) for vv in v)
    if isinstance(v, dict):
        return '{' + ', '.join(f'{k}: {format(vv)}' for k, vv in v.items()) + '}'
    return str(v)

def adaptive_rule(df):
    timespan = (df.index[-1] - df.index[0]).total_seconds()
    if timespan < 600:
        return '15s'
    elif timespan < 7200:
        return '1min'
    else:
        return '10min'
    
class Reader:

    def __init__(self, run_name, prefix=''):
        self._reader = numpy.Reader(run_name, 'stats')
        self._prefix = prefix
        self._arrs = {}

    def arrays(self):
        #TODO: If this gets slow, do amortized allocation of arrays x2 as big as needed
        for channel, new in self._reader.read().items():
            category, field = re.match(r'^(.*?)/(.*)$', channel).groups()
            if field.startswith(self._prefix):
                current = [self._arrs[category, field]] if (category, field) in self._arrs else []
                self._arrs[category, field] = np.concatenate(current + new)
        return arrdict.arrdict(self._arrs)

    def pandas(self):
        arrs = self.arrays()

        dfs = {}
        for (category, field), arr in arrs.items():
            df = pd.DataFrame.from_records(arr, index='_time')
            df.index.name = 'time'
            dfs[category, field] = df
        return arrdict.arrdict(dfs)
        
    def resample(self, rule='60s', **kwargs):
        kwargs = {'rule': rule, **kwargs}

        results = {}
        for (category, field), df in self.pandas().items():
            func = getattr(categories, category)
            results[field] = func(**{k: df[k] for k in df})(**kwargs)

        if results:
            df = pd.concat(results, 1)
            df.index = df.index - df.index[0]
            return df
        else:
            return pd.DataFrame(index=pd.TimedeltaIndex([], name='time'))

def arrays(prefix='', run_name=-1):
    return Reader(run_name, prefix).arrays()

def pandas(name, run_name=-1):
    dfs = Reader(run_name, name).pandas()
    for (_, field), df in dfs.items():
        return df
    raise KeyError(f'Couldn\'t find a statistic matching {name}')

def resample(prefix='', run_name=-1, rule='60s'):
    return Reader(run_name, prefix).resample(rule)

def tdformat(td):
    """How is this not in Python, numpy or pandas?"""
    x = td.total_seconds()
    x, _ = divmod(x, 1)
    x, s = divmod(x, 60)
    if x < 1:
        return f'{s:.0f}s'
    h, m = divmod(x, 60)
    if h < 1:
        return f'{m:.0f}m{s:02.0f}s'
    else:
        return f'{h:.0f}h{m:02.0f}m{s:02.0f}s'

def __from_dir(canceller, run_name, out, rule, throttle=1):
    reader = Reader(run_name)
    start = pd.Timestamp.now()

    nxt = time.time()
    while True:
        if time.time() > nxt:
            nxt = nxt + throttle

            # Base slightly into the future, else by the time the resample actually happens you're 
            # left with an almost-empty last interval.
            base = int(time.time() % 60) + 5
            values = reader.resample(rule=rule, base=base)
            
            if len(values) > 0:
                values = values.ffill(limit=1).iloc[-1].to_dict()
                key_length = max([len(str(k)) for k in values], default=0)+1
                content = '\n'.join(f'{{:{key_length}s}} {{}}'.format(k, format(values[k])) for k in sorted(values))
            else:
                content = 'No stats yet'

            size = paths.size(run_name, 'stats')
            age = pd.Timestamp.now() - start
            out.refresh(f'{run_name}: {tdformat(age)} old, {rule} rule, {size:.0f}MB on disk\n\n{content}')

        if canceller.is_set():
            break

        time.sleep(.1)

def _from_dir(*args, **kwargs):
    try:
        __from_dir(*args, **kwargs)
    except KeyboardInterrupt:
        log.info('Interrupting main')
        _thread.interrupt_main()

@contextmanager
def from_dir(run_name, compositor=None, rule='60s'):
    if logging.in_ipython():
        try:
            canceller = threading.Event()
            out = (compositor or widgets.Compositor()).output()
            thread = threading.Thread(target=_from_dir, args=(canceller, run_name, out, rule))
            thread.start()
            yield
        finally:
            canceller.set()
            thread.join(1)
            if thread.is_alive():
                log.error('Stat display thread won\'t die')
            else:
                log.info('Stat display thread cancelled')

            # Want to leave the outputs open so you can see the final stats
            # out.close()
    else:
        log.info('No stats emitted in console mode')
        yield


import time
import torch
from torch import multiprocessing as mp
import queue
from contextlib import contextmanager, asynccontextmanager
import traceback
import asyncio
from functools import wraps
from torch import nn
import logging
from . import dotdict

log = logging.getLogger(__name__)

class SerialQueue:

    def __init__(self):
        self._queue = []
        self._put_end = False
        self._got_end = False

    def get(self):
        if len(self._queue) > 0:
            item = self._queue[0]
            self._queue = self._queue[1:]

            if isinstance(item, str) and (item == '__END__'):
                log.info('Got END')
                self._got_end = True
                return None
            else:
                return item
        else:
            return None
    
    def put(self, item):
        # Not safe to test `in` directly since `item` is likely to be a tensor
        if isinstance(item, (str, type(None))) and (item in ('__END__', None)):
            raise ValueError(f'Tried to put sentinel value "{item}"')
        if len(self._queue) < 1:
            self._queue.append(item)
            return True
        else:
            return False

    def put_end(self):
        if self._put_end:
            return True
        else:
            if len(self._queue) < 1:
                self._queue.append('__END__')
                log.info('Put END')
                self._put_end = True
                return True
            else:
                return False

    def get_end(self):
        self.get()
        return self._got_end
        
    def join(self, timeout=None):
        if len(self._queue) == 0:
            return True
        else:
            return False

class MultiprocessQueue:

    def __init__(self):
        self.queue = mp.JoinableQueue(1)
        self._put_end = False
        self._got_end = False

    def get(self):
        try:
            item = self.queue.get_nowait()
            if isinstance(item, str) and (item == '__END__'):
                log.info('Got END')
                self._got_end = True
                self.queue.task_done()
                return None
            else:
                self.queue.task_done()
                return item
        except queue.Empty:
            return None
    
    def put(self, item):
        # Not safe to test `in` directly since `item` is likely to be a tensor
        if isinstance(item, (str, type(None))) and (item in ('__END__', None)):
            raise ValueError(f'Tried to put sentinel value "{item}"')
        try:
            self.queue.put_nowait(item)
            return True
        except queue.Full:
            return False

    def put_end(self):
        try:
            if not self._put_end:
                self.queue.put_nowait('__END__')
                log.info('Put END')
                self._put_end = True
            return True
        except queue.Full:
            return False

    def get_end(self):
        self.get()
        return self._got_end
        
    def join(self, timeout=None):
        try:
            with self.queue._cond:
                if not self.queue._unfinished_tasks._semlock._is_zero():
                    self.queue._cond.wait(timeout=timeout)
            return True
        except RuntimeError:
            return False

async def close(intakes, outputs, timeout=5):
    """Strategy:
        * Wait until you can send an END through each output queue
        * Drain the intake queues until you get an END from each one
        * Wait for each output queue to drain
    """
    log.info(f'Closing; draining intakes and waiting to send ENDs. {timeout}s timeout.')
    cutoff = time.time() + timeout
    while True:
        # Avoid a deadlock where everyone's queues are full so ENDs can't be sent
        for intake in intakes:
            intake.get()
        
        if all(o.put_end() for o in outputs):
            break
        if time.time() > cutoff:
            log.warn('Timed out while waiting to send ENDs')
            return 

        # We're not actually running in a proper scheduler here, so can't sleep via it
        await asyncio.sleep(0)
        time.sleep(.1)
    
    log.info(f'Sent ENDs to outputs; waiting to get ENDs from intakes')
    while True:
        if all(i.get_end() for i in intakes):
            break
        if time.time() > cutoff:
            log.warn('Timed out while waiting to get ENDs')
            return

        # We're not actually running in a proper scheduler here, so can't sleep via it
        await asyncio.sleep(0)
        time.sleep(.1)

    log.info(f'Intakes emptied; waiting for outputs to drain')
    while True:
        if all(o.join(.1) for o in outputs):
            break
        if time.time() > cutoff:
            log.warn('Timed out while waiting to drain outputs')
            return

        # We're not actually running in a proper scheduler here, so can't sleep via it
        await asyncio.sleep(0)
        time.sleep(.1)

    log.info('Outputs drained.')

def create(x, serial=False):
    if isinstance(x, dict):
        return dotdict.dotdict({n: create(v, serial) for n, v in x.items()})
    elif isinstance(x, (list, tuple)):
        return dotdict.dotdict({n: create(n, serial) for n in x})
    elif isinstance(x, str):
        return SerialQueue() if serial else MultiprocessQueue()
    raise ValueError(f'Can\'t handle {type(x)}')

@asynccontextmanager
async def cleanup(intakes, outputs):
    intakes = [intakes] if isinstance(intakes, (SerialQueue, MultiprocessQueue)) else intakes
    outputs = [outputs] if isinstance(outputs, (SerialQueue, MultiprocessQueue)) else outputs
    try:
        yield
    except:
        log.info(f'Got an exception, cleaning up queues:\n{traceback.format_exc()}')
        raise
    finally:
        await close(intakes, outputs)




import ipywidgets as widgets
from IPython.display import display, clear_output
import threading

WRITE_LOCK = threading.RLock()

class Output:

    def __init__(self, compositor, output, lines):
        self._compositor = compositor
        self._output = output
        self.lines = lines

    def refresh(self, content):
        # This is not thread-safe, but the recommended way to do 
        # thread-safeness - to use append_stdout - causes flickering
        with WRITE_LOCK, self._output:
            clear_output(wait=True)
            print(content)
    
    def close(self):
        self._compositor.remove(self._output)

class Compositor:

    def __init__(self, lines=80):
        self.lines = lines
        self._box = widgets.HBox(
            layout=widgets.Layout(align_items='stretch'))
        display(self._box)

    def output(self):
        output = widgets.Output(
            layout=widgets.Layout(width='100%'))
        self._box.children = (*self._box.children, output)

        return Output(self, output, self.lines)

    def remove(self, child):
        child.close()
        self._box.children = tuple(c for c in self._box.children if c != child)

    def clear(self):
        for child in self._box.children:
            self.remove(child)


def test():
    compositor = Compositor()
    first = compositor.output()
    second = compositor.output()

    first.refresh('left')
    second.refresh('right')


from rebar import arrdict, dotdict
from megastep import modules, core, toys, scene
from torch import nn
from megastep.demo import heads
import matplotlib.pyplot as plt

class Minimal:

    def __init__(self, n_envs=1):
        """A minimal environment, with a box env, depth observations and simple movement. A good foundation for
        building your own environments.
        
        See :ref:`the simple environment tutorial for details <minimal-env>`."""

        geometries = n_envs*[toys.box()]
        scenery = scene.scenery(geometries, n_agents=1)
        self.core = core.Core(scenery)
        self.spawner = modules.RandomSpawns(geometries, self.core)
        self.rgb = modules.RGB(self.core)
        self.movement = modules.SimpleMovement(self.core)

        self.obs_space = self.rgb.space
        self.action_space = self.movement.space

    def reset(self):
        self.spawner(self.core.agent_full(True))
        return arrdict.arrdict(obs=self.rgb())

    def step(self, decision):
        self.movement(decision)
        return arrdict.arrdict(obs=self.rgb())

    def state(self, e=0):
        return dotdict.dotdict(
            core=self.core.state(e),
            rgb=self.rgb.state(e))
    
    @classmethod
    def plot_state(self, state):
        fig = plt.figure()
        gs = plt.GridSpec(1, 3, fig)

        plan = plt.subplot(gs[:, :2])
        core.Core.plot_state(state.core, plan)

        im = plt.subplot(gs[:, -1])
        modules.RGB.plot_state(state.rgb, [im])

        return fig

class Agent(nn.Module):
    """A minimal agent to go with the minimal environment.

    See :ref:`the simple environment tutorial for details <minimal-env>`.
    """

    def __init__(self, env, width=32):
        super().__init__()
        self.intake = heads.intake(env.obs_space, width)
        self.output = heads.output(env.action_space, width)
        self.policy = nn.Sequential(self.intake, self.output)
        
    def forward(self, world):
        logits = self.policy(world.obs)
        actions = self.output.sample(logits)
        return arrdict.arrdict(logits=logits, actions=actions)



import numpy as np
from numpy.lib import format as npformat
from . import paths
from io import BytesIO
from datetime import datetime
from collections import defaultdict
import time

def infer_dtype(exemplar):
    return np.dtype([(k, v.dtype if isinstance(v, np.generic) else type(v)) for k, v in exemplar.items()])

def make_header(dtype):
    """
    Ref: https://numpy.org/devdocs/reference/generated/numpy.lib.format.html
    We're doing version 3. Only difference is the zero shape, since we're
    going to deduce the array size from the filesize.
    """
    assert not dtype.hasobject, 'Arrays with objects in get pickled, so can\'t be appended to'

    bs = BytesIO()
    npformat._write_array_header(bs, {
        'descr': dtype.descr, 
        'fortran_order': False, 
        'shape': (0,)},
        version=(3, 0))
    return bs.getvalue()

class FileWriter:

    def __init__(self, path, period=1):
        self._path = path
        self._file = None
        self._period = 5
        self._next = time.time()
        
    def _init(self, exemplar):
        self._file = self._path.open('wb', buffering=4096)
        self._dtype = infer_dtype(exemplar)
        self._file.write(make_header(self._dtype))
        self._file.flush()

    def write(self, d):
        if self._file is None:
            self._init(d)
        assert set(d) == set(self._dtype.names)
        row = np.array([tuple(v for v in d.values())], self._dtype)
        self._file.write(row.tobytes())
        self._file.flush()

    def close(self):
        self._file.close()
        self._file = None

class Writer:

    def __init__(self, run_name, group):
        self._run_name = run_name
        self._group = group
        self._writers = {}

    def write(self, channel, d):
        if channel not in self._writers:
            path = paths.path(self._run_name, self._group, channel).with_suffix('.npr')
            self._writers[channel] = FileWriter(path)
        self._writers[channel].write(d)

    def write_many(self, ds):
        for channel, d in ds.items():
            if channel not in self._writers:
                path = paths.path(self._run_name, self._group, channel).with_suffix('.npr')
                self._writers[channel] = FileWriter(path)
            self._writers[channel].write(d)

    def close(self):
        for _, w in self._writers.items():
            w.close()
        self._writers = {}

class FileReader:

    def __init__(self, path):
        self._path = path
        self._file = None

    def _init(self):
        #TODO: Can speed this up with PAG's regex header parser
        self._file = self._path.open('rb')
        version = npformat.read_magic(self._file)
        _, _, dtype = npformat._read_array_header(self._file, version)
        self._dtype = dtype

    def read(self):
        if self._file is None:
            self._init()
        return np.fromfile(self._file, dtype=self._dtype)

    def close(self):
        self._file.close()
        self._file = None

class Reader:

    def __init__(self, run_name, group):
        self._run_name = paths.resolve(run_name)
        self._group = group
        self._readers = {}

    def read(self):
        for path in paths.subdirectory(self._run_name, self._group).glob('**/*.npr'):
            parts = paths.parse(path)
            if (parts.channel, parts.filename) not in self._readers:
                self._readers[parts.channel, parts.filename] = FileReader(path)

        results = defaultdict(lambda: [])
        for (channel, _), reader in self._readers.items():
            arr = reader.read()
            if len(arr) > 0:
                results[channel].append(arr)

        return results


def test_file_write_read():
    d = {'total': 65536, 'count': 14, '_time': np.datetime64('now')}
    
    paths.clear('test', 'stats')
    path = paths.path('test', 'stats', 'mean/traj-length').with_suffix('.npr')

    writer = FileWriter(path)
    writer.write(d)

    reader = FileReader(path)
    r = reader.read()

    assert len(r) == 1

def test_write_read():
    paths.clear('test', 'stats')

    writer = Writer('test', 'stats')
    writer.write('mean/traj-length', {'total': 65536, 'count': 14, '_time': np.datetime64('now')})
    writer.write('max/reward', {'total': 50000.5, 'count': 50, '_time': np.datetime64('now')})

    reader = Reader('test', 'stats')
    r = reader.read()

    assert len(r) == 2


"""TODO-DOCS Explorer docs"""
import numpy as np
import torch
from megastep import modules, core, plotting, scene, cubicasa
from rebar import arrdict, dotdict
import matplotlib.pyplot as plt

class Explorer: 

    def __init__(self, n_envs, *args, **kwargs):
        geometries = cubicasa.sample(n_envs)
        scenery = scene.scenery(geometries, 1)
        self.core = core.Core(scenery, *args, res=4*64, fov=130, **kwargs)
        self._rgb = modules.RGB(self.core, n_agents=1, subsample=4)
        self._depth = modules.Depth(self.core, n_agents=1, subsample=4)
        self._mover = modules.MomentumMovement(self.core)
        self._imu = modules.IMU(self.core)
        self._respawner = modules.RandomSpawns(geometries, self.core)

        self.action_space = self._mover.space
        self.obs_space = dotdict.dotdict(
            rgb=self._rgb.space,
            d=self._depth.space,
            imu=self._imu.space)

        self._tex_to_env = self.core.scenery.lines.inverse[self.core.scenery.textures.inverse.long()].long()
        self._seen = torch.full_like(self._tex_to_env, False)
        self._potential = self.core.env_full(0.)

        self._lengths = torch.zeros(self.core.n_envs, device=self.core.device, dtype=torch.int)

        self.device = self.core.device

    def _tex_indices(self, aux): 
        scenery = self.core.scenery 
        mask = aux.indices >= 0
        result = torch.full_like(aux.indices, -1, dtype=torch.long)
        tex_n = (scenery.lines.starts[:, None, None, None] + aux.indices)[mask]
        tex_w = scenery.textures.widths[tex_n.to(torch.long)]
        tex_i = torch.min(torch.floor(tex_w.to(torch.float)*aux.locations[mask]), tex_w.to(torch.float)-1)
        tex_s = scenery.textures.starts[tex_n.to(torch.long)]
        result[mask] = tex_s.to(torch.long) + tex_i.to(torch.long)
        return result.unsqueeze(2)

    def _reward(self, r, reset):
        texindices = self._tex_indices(r)
        self._seen[texindices] = True

        potential = torch.zeros_like(self._potential)
        potential.scatter_add_(0, self._tex_to_env, self._seen.float())

        reward = (potential - self._potential)/(self.core.res//self._rgb.subsample)
        self._potential = potential

        # Should I render twice so that the last reward is accurate?
        reward[reset] = 0.

        return reward

    def _observe(self, reset):
        r = modules.render(self.core)
        obs = arrdict.arrdict(
                rgb=self._rgb(r), 
                d=self._depth(r), 
                imu=self._imu())
        reward = self._reward(r, reset)
        return obs, reward

    def _reset(self, reset=None):
        self._respawner(reset.unsqueeze(-1))
        self._seen[reset[self._tex_to_env]] = False
        self._potential[reset] = 0
        self._lengths[reset] = 0

    @torch.no_grad()
    def reset(self):
        reset = self.core.env_full(True)
        self._reset(reset)
        obs, reward = self._observe(reset)
        return arrdict.arrdict(
            obs=obs,
            reset=reset, 
            reward=reward)

    @torch.no_grad()
    def step(self, decision):
        self._mover(decision)

        self._lengths += 1

        reset = (self._lengths >= self._potential + 200)
        self._reset(reset)
        obs, reward = self._observe(reset)
        return arrdict.arrdict(
            obs=obs,
            reset=reset, 
            reward=reward) 

    def state(self, e=0):
        seen = self._seen[self._tex_to_env == e]
        return arrdict.arrdict(
            core=self.core.state(e),
            rgb=self._rgb.state(e),
            d=self._depth.state(e),
            potential=self._potential[e].clone(),
            seen=seen.clone(),
            length=self._lengths[e].clone(),
            max_length=self._potential[e].add(200).clone())

    @classmethod
    def plot_state(cls, state):
        fig = plt.figure()
        gs = plt.GridSpec(2, 2, fig, 0, 0, 1, 1)

        alpha = .1 + .9*state.seen.astype(float)
        # modifying this in place will bite me eventually. o for a lens
        state.core.scenery.textures.vals = np.concatenate([state.core.scenery.textures.vals, alpha[:, None]], 1)
        ax = core.Core.plot_state(state.core, plt.subplot(gs[:, 0]))

        images = {'rgb': state.rgb, 'd': state.d}
        plotting.plot_images(images, [plt.subplot(gs[:, 1])])

        s = (f'length: {state.length:d}/{state.max_length:.0f}\n'
            f'potential: {state.potential:.0f}')
        ax.annotate(s, (5., 5.), xycoords='axes points')

        return fig

    def display(self, d=0):
        return self.plot_state(arrdict.numpyify(self.state(d)))

